/* 
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.techfak.scie.pdfextractor;

import java.awt.Container;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.HeadlessException;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.concurrent.locks.ReentrantLock;
import javax.swing.DefaultListModel;
import javax.swing.DefaultListSelectionModel;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JProgressBar;
import javax.swing.JScrollPane;

/**
 * This is the commandline/executable interface of the extractor. It starts the
 * graphical user
 * interface and takes no parameters at all.
 *
 */
public class App {

	private static final String TITLE = "SCIE PDF Text Extractor";
	private static final String VERSION = "1.0";
	private static final MainWindow MAINWINDOW = new MainWindow();
	private static DefaultListModel<File> input = new DefaultListModel<>();
	private static File output = new File(".");

	public static void main(String[] args) {
		MAINWINDOW.start();
	}

	private static void process() {
		final HashSet<File> actualFiles = new HashSet<>();
		final Enumeration<File> files = input.elements();
		//first get all unique files.
		while (files.hasMoreElements()) {
			final File file = files.nextElement();
			if (file.isDirectory()) {
				for (final File childFile : file.listFiles()) {
					if (childFile.getName().endsWith(".pdf")) {
						actualFiles.add(childFile);
					}
				}
			} else {
				actualFiles.add(file);
			}
		}
		//then start actual calculations.
		float counter = 0;
		final float total = actualFiles.size();
		for (final File file : actualFiles) {
			processPDFFile(file);
			counter++;
			final float percent = counter / total * 100;
			MAINWINDOW.setProgress((int) percent);
		}
		MAINWINDOW.start();
		JOptionPane.showMessageDialog(MAINWINDOW, "Bearbeitung abgeschlossen!",
				"Finished!", JOptionPane.INFORMATION_MESSAGE);
	}

	private static void processPDFFile(final File file) {
		try {
			//at this point we now that we are treating an actual pdf.
			final FileInputStream input = new FileInputStream(file);
			final PDFStructuredTextExtractor ext = new PDFStructuredTextExtractor(input);
			final String docText = ext.doImport().indexedToString(0);
			final String outFileName = file.getName().replaceAll("\\.pdf$", ".txt");
			final File outFile = new File(output, outFileName);
			outFile.delete();
			final BufferedWriter out = new BufferedWriter(new FileWriter(outFile));
			out.write(docText);
			out.close();
			input.close();
		} catch (IOException ex) {
			MAINWINDOW.showError("Error during parsing of "
					+ file.getAbsolutePath() + ":\n" + ex.getMessage());
//			System.exit(1);
		}
	}

	private static class MainWindow extends JFrame {

		private final JLabel outputLabel = new JLabel(new File(".").getAbsolutePath());

		public MainWindow() throws HeadlessException {
			super(TITLE + " Version " + VERSION);
			setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		}

		public void start() {
			final Container pane = getContentPane();
			pane.removeAll();
			pane.setPreferredSize(new Dimension(800, 400));
			pane.setLayout(new GridBagLayout());
			final GridBagConstraints c = new GridBagConstraints();
			c.insets = new Insets(2, 15, 2, 15);
			c.fill = GridBagConstraints.HORIZONTAL;
			//Input file chooser.
			c.weightx = 0.5;
			c.weighty = 0.1;
			pane.add(new JLabel("Input Dateien: "), c);
			c.gridx = 1;
			final JButton inputButton = new JButton("Auswählen");
			pane.add(inputButton, c);
			inputButton.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					final JFileChooser chooser = new JFileChooser(".");
					chooser.setMultiSelectionEnabled(true);
					chooser.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
					final int returnVal = chooser.showOpenDialog(MAINWINDOW);
					if (returnVal == JFileChooser.APPROVE_OPTION) {
						final File[] files = chooser.getSelectedFiles();
						boolean valid = true;
						for (final File file : files) {
							if (file.isDirectory() || file.getName().endsWith(".pdf")) {
								input.addElement(file);
							} else {
								valid = false;
							}
						}
						if (!valid) {
							JOptionPane.showMessageDialog(MAINWINDOW,
									"Bitte wählen Sie .pdf-Dateien oder Ordner aus.",
									"Error", JOptionPane.ERROR_MESSAGE);
						}
					}
				}
			});
			//input file list.
			c.gridx = 0;
			c.gridy = 1;
			c.weightx = 1;
			c.weighty = 0.6;
			c.gridwidth = 2;
			final JList<File> inputListView = new JList<>(input);
			final JScrollPane scrollPane = new JScrollPane(inputListView);
			inputListView.setOpaque(false);
			inputListView.setSelectionModel(new DisablingListSelectionModel());
			pane.add(scrollPane, c);
			//output folder chooser.
			c.gridx = 0;
			c.gridy = 2;
			c.gridwidth = 1;
			c.weightx = 0.5;
			c.weighty = 0.1;
			pane.add(new JLabel("Output Ordner: "), c);
			c.gridx = 1;
			final JButton outputButton = new JButton("Auswählen");
			pane.add(outputButton, c);
			outputButton.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					final JFileChooser chooser = new JFileChooser(".");
					chooser.setMultiSelectionEnabled(false);
					chooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
					final int returnVal = chooser.showSaveDialog(MAINWINDOW);
					if (returnVal == JFileChooser.APPROVE_OPTION) {
						output = chooser.getSelectedFile();
						outputLabel.setText(output.getAbsolutePath());
						MAINWINDOW.repaint();
					}
				}
			});
			//output name.
			c.gridx = 0;
			c.gridy = 3;
			c.gridwidth = 2;
			c.weightx = 1;
			pane.add(outputLabel, c);
			//start button.
			c.gridx = 0;
			c.gridy = 4;
			c.gridwidth = 2;
			c.weightx = 1;
			c.fill = GridBagConstraints.CENTER;
			final JButton startButton = new JButton("Start");
			pane.add(startButton, c);
			startButton.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					progressInit();
					try {
						Thread.sleep(1500);
					} catch (InterruptedException ex) {
						throw new RuntimeException(ex);
					}
					final Runnable procRunnable = new Runnable() {
						@Override
						public void run() {
							process();
						}
					};
					final Thread procThread = new Thread(procRunnable);
					procThread.start();
				}
			});
			//finalizing
			pack();
			setVisible(true);
		}
		private final JProgressBar progBar = new JProgressBar(0, 100);

		public void progressInit() {
			final Container pane = getContentPane();
			pane.removeAll();
			pane.setPreferredSize(new Dimension(800, 400));
			pane.setLayout(new GridBagLayout());
			final GridBagConstraints c = new GridBagConstraints();
			c.weightx = 1;
			c.weighty = 1;
			c.anchor = GridBagConstraints.CENTER;
			pane.add(progBar, c);
			progBar.setString("0 % abgeschlossen");
			progBar.setStringPainted(true);
			setContentPane(pane);
			repaint();
			//finalizing
			pack();
			setVisible(true);
		}
		private final ReentrantLock progLock = new ReentrantLock();

		public void setProgress(float percent) {
			progLock.lock();
			try {
				final int newVal = (int) percent;
				progBar.setValue(newVal);
				progBar.setString(newVal + " % abgeschlossen");
			} finally {
				progLock.unlock();
			}
		}

		public void showError(String text) {
			JOptionPane.showMessageDialog(MAINWINDOW,
					text, "Error", JOptionPane.ERROR_MESSAGE);
		}
	}

	private static class DisablingListSelectionModel extends DefaultListSelectionModel {

		@Override
		public void setSelectionInterval(int index0, int index1) {
			super.setSelectionInterval(-1, -1); //remove any selection.
		}
	}
}
